/*
    Voxel Structure based on Superscalar by Bobcao3
*/

const int volumeWidth = 256;
const int volumeHeight = 256;

const int voxelmapResolution = voxelMapRes;
const ivec2 voxelSegmentEdge = ivec2(voxelmapResolution, voxelmapResolution / 2) - 1;

const int volumeDepthGridSize = (voxelmapResolution / volumeWidth);

const int volumeDepth = ((voxelmapResolution / volumeWidth) * (voxelmapResolution / volumeHeight)) / 2;

ivec3 sceneToVoxelIndex(vec3 world_pos) {
    return ivec3(floor(world_pos + mod(cameraPosition, 1.0))) + ivec3(volumeWidth, volumeDepth, volumeHeight) / 2;
}
ivec3 sceneToVoxelIndexNoCam(vec3 world_pos) {
    return ivec3(floor(world_pos)) + ivec3(volumeWidth, volumeDepth, volumeHeight) / 2;
}
vec3 voxelToSceneSpace(ivec3 voxelPos) {
    return vec3(voxelPos) - vec3(volumeWidth, volumeDepth, volumeHeight) / 2.0;
}
vec3 voxelToSceneSpace(vec3 voxelPos) {
    return vec3(voxelPos) - vec3(volumeWidth, volumeDepth, volumeHeight) / 2.0 - mod(cameraPosition, 1.0);
}
vec3 sceneToVoxelSpace(vec3 world_pos) {
    return (world_pos + mod(cameraPosition, 1.0)) + vec3(volumeWidth, volumeDepth, volumeHeight) / 2.0;
}
ivec3 voxelSpaceToIndex(vec3 voxelPos) {
    return ivec3(floor(voxelPos));
}

bool outsideVoxelVolume(ivec3 pos) {
    return pos.x < 0 || pos.x >= volumeWidth || pos.y < 0 || pos.y >= volumeDepth || pos.z < 0 || pos.z >= volumeHeight;
}
bool outsideVoxelVolume(vec3 pos) {
    return pos.x < 0 || pos.x >= volumeWidth || pos.y < 0 || pos.y >= volumeDepth || pos.z < 0 || pos.z >= volumeHeight;
}
bool sceneWithinVoxelVolume(vec3 pos) {
    return !outsideVoxelVolume(sceneToVoxelIndexNoCam(pos));
}

ivec2 getVoxelPixel(ivec3 pos) {
    if (outsideVoxelVolume(pos)) return ivec2(-1);

    ivec2 base = ivec2((pos.y % volumeDepthGridSize) * volumeWidth, (pos.y / volumeDepthGridSize) * volumeHeight);
    ivec2 uv = base + pos.xz;

    return uv;
}

vec2 getVoxelUV(vec3 pos) {
    if (outsideVoxelVolume(pos)) return ivec2(-1);

    ivec2 base = ivec2((int(pos.y) % volumeDepthGridSize) * volumeWidth, (int(pos.y) / volumeDepthGridSize) * volumeHeight);
    vec2 uv = vec2(base) + pos.xz;

    return uv;
}

ivec3 uvToVoxelIndex(ivec2 uv) {
    ivec2 y_grid = uv / ivec2(volumeWidth, volumeHeight);
    ivec2 xz_pos = uv % ivec2(volumeWidth, volumeHeight);
    ivec3 wpos = ivec3(xz_pos.x, y_grid.x + y_grid.y * volumeDepthGridSize, xz_pos.y);

    return wpos;
}