#pragma optionNV(unroll all)

#include "/lib/head.glsl"
#include "/lib/util/encoders.glsl"

layout (triangles) in;

layout (triangle_strip, max_vertices = 6) out;
const int maxVerticesOut = 6;

in vertexOut {
    flat int voxelMode;
    flat int blockID;
    flat int emissionID;

    vec2 UV;
    vec2 MidUV;

    vec3 Tint;
    vec3 LocalLocation;
    vec3 SceneLocation;
    vec3 CenteredSceneLocation;
    flat vec3 Normal;

    flat vec4 Tangent;
} vertexIn[3];


uniform vec3 cameraPosition;

uniform mat4 shadowModelViewInverse;

#include "/lib/light/warp.glsl"

out voxelOut {
    flat int isVoxel;
    flat int emission;
    flat int voxelType;

    vec2 midUV;
    vec2 minUV;
    vec2 maxUV;
};

out geoOut {
    flat int matID;

    float warp;

    vec2 UV;

    vec3 tint;
    vec3 scenePos;
};

#include "/lib/voxel/store.glsl"

void main() {
    for (int i = 0; i < 3; ++i) {
        isVoxel = 0;
        tint    = vertexIn[i].Tint;
        UV      = vertexIn[i].UV;
        matID   = vertexIn[i].blockID;

        gl_Position     = gl_ModelViewMatrix * vec4(vertexIn[i].LocalLocation, 1.0);

        scenePos        = vertexIn[i].SceneLocation;

        gl_Position     = gl_ProjectionMatrix * gl_Position;
        gl_Position.xy  = shadowmapWarp(gl_Position.xy, warp);
        gl_Position.z  *= 0.2;
        gl_Position.x   = gl_Position.x * 0.5 + 0.5;
        gl_Position.xy  = gl_Position.yx;
        if (gl_Position.y <= 0.0) return;

        EmitVertex();
    } EndPrimitive();

    if (vertexIn[0].voxelMode == 0) return;

    vec2 midCoord   = vertexIn[0].MidUV;
    vec2 minCoord   = min(min(vertexIn[0].UV, vertexIn[1].UV), vertexIn[2].UV);
    vec2 maxCoord   = max(max(vertexIn[0].UV, vertexIn[1].UV), vertexIn[2].UV);

    vec3 avgTint    = (vertexIn[0].Tint + vertexIn[1].Tint + vertexIn[2].Tint) / 3.0;

    vec3 normal     = mat3(shadowModelViewInverse) * normalize(gl_NormalMatrix * vertexIn[0].Normal);

    float normalBias = 0.5 - 0.25 * normal.y;

    vec3 triangleBoundsMin  = min(vertexIn[0].CenteredSceneLocation, min(vertexIn[1].CenteredSceneLocation, vertexIn[2].CenteredSceneLocation));
    vec3 triangleBoundsMax  = max(vertexIn[0].CenteredSceneLocation, max(vertexIn[1].CenteredSceneLocation, vertexIn[2].CenteredSceneLocation));

    vec3 triangleMidLocation = (triangleBoundsMin + triangleBoundsMax) * 0.5;

    ivec3 voxelIndex     = sceneToVoxelIndex(triangleMidLocation);

    int isEmissive      = vertexIn[0].emissionID;

    if (!outsideVoxelVolume(voxelIndex)) {
        vec2 texelCoord     = vec2(getVoxelPixel(voxelIndex)) / voxelmapResolution;
            texelCoord      = texelCoord * 2.0 - 1.0;

        const vec2[3] verts = vec2[3](vec2(1,-1) / shadowMapRes, vec2(0,1) / shadowMapRes, vec2(-1,-1) / shadowMapRes);

        for (int j = 0; j < 3; ++j) {
            isVoxel         = 1;
            emission        = isEmissive;
            midUV           = midCoord;
            minUV           = minCoord;
            maxUV           = maxCoord;
            tint            = avgTint;
            voxelType       = vertexIn[0].voxelMode;

            gl_Position     = vec4(texelCoord, normalBias, 1.0);
            gl_Position.xy += verts[j];

            EmitVertex();
        }
        EndPrimitive();
    } else {
        return;
    }

}